"""Storage adapter interfaces."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Mapping, Optional


@dataclass
class StorageEntry:
    content: str
    version: int = 0


class StorageAdapter:
    def read(self, path: str) -> StorageEntry:
        raise NotImplementedError

    def write(self, path: str, content: str, *, version: int) -> StorageEntry:
        raise NotImplementedError

    def list(self) -> Dict[str, StorageEntry]:
        raise NotImplementedError

    def delete(self, path: str) -> None:
        raise NotImplementedError


@dataclass
class MemoryStorageAdapter(StorageAdapter):
    initial: Mapping[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self._entries: Dict[str, StorageEntry] = {
            path: StorageEntry(content=text, version=0)
            for path, text in self.initial.items()
        }

    def read(self, path: str) -> StorageEntry:
        entry = self._entries.get(path)
        if entry is None:
            raise FileNotFoundError(path)
        return StorageEntry(content=entry.content, version=entry.version)

    def write(self, path: str, content: str, *, version: int) -> StorageEntry:
        entry = self._entries.get(path)
        if entry and entry.version != version:
            raise ValueError("version mismatch")
        next_version = version + 1
        entry = StorageEntry(content=content, version=next_version)
        self._entries[path] = entry
        return StorageEntry(content=entry.content, version=entry.version)

    def list(self) -> Dict[str, StorageEntry]:
        return {path: StorageEntry(content=entry.content, version=entry.version) for path, entry in self._entries.items()}

    def delete(self, path: str) -> None:
        self._entries.pop(path, None)



from sandfs import VirtualFileSystem
from sandfs.integrations import PathEvent


def test_path_hook_receives_create_update_delete():
    vfs = VirtualFileSystem()
    events: list[PathEvent] = []

    def handler(event: PathEvent) -> None:
        events.append(event)

    vfs.register_path_hook("/blue/inbox", handler)

    vfs.write_file("/blue/inbox/note.txt", "hello")
    vfs.write_file("/blue/inbox/note.txt", "world")
    vfs.remove("/blue/inbox/note.txt")

    assert [e.event for e in events] == ["create", "update", "delete"]
    assert events[0].content == "hello"
    assert events[1].content == "world"
    assert events[2].content is None



"""Policy and visibility helpers for sandfs."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import FrozenSet, Iterable, Optional, Set


@dataclass
class NodePolicy:
    """Controls access, write semantics, and visibility for a node."""

    readable: bool = True
    writable: bool = True
    append_only: bool = False
    classification: str = "public"
    principals: Set[str] = field(default_factory=set)


@dataclass(frozen=True)
class VisibilityView:
    """Filters nodes by classification labels and principals."""

    classifications: Optional[FrozenSet[str]] = None
    principals: Optional[FrozenSet[str]] = None

    def __init__(
        self,
        classifications: Optional[Iterable[str]] = None,
        principals: Optional[Iterable[str]] = None,
    ) -> None:
        object.__setattr__(
            self,
            "classifications",
            frozenset(classifications) if classifications is not None else None,
        )
        object.__setattr__(
            self,
            "principals",
            frozenset(principals) if principals is not None else None,
        )

    def allows(self, policy: NodePolicy) -> bool:
        if policy.principals:
            if self.principals is None:
                return False
            if not (policy.principals & self.principals):
                return False
            return True
        if self.classifications is None:
            return True
        return policy.classification in self.classifications


__all__ = ["NodePolicy", "VisibilityView"]



from sandfs import VirtualFileSystem
from sandfs.providers import ProvidedNode


def test_write_and_read_file():
    vfs = VirtualFileSystem()
    vfs.write_file("/notes/todo.txt", "- build VFS\n")
    assert vfs.read_file("/notes/todo.txt") == "- build VFS\n"
    entries = vfs.ls("/notes")
    assert entries[0].name == "todo.txt"


def test_dynamic_file_provider():
    vfs = VirtualFileSystem()

    def provider(ctx):
        return f"dynamic at {ctx.path}"

    vfs.mount_file("/dynamic/info.txt", provider)
    assert "dynamic at /dynamic/info.txt" == vfs.read_file("/dynamic/info.txt")


def test_directory_provider_populates_children():
    vfs = VirtualFileSystem()

    def loader(ctx):
        return {
            "README.md": ProvidedNode.file(content="hello"),
            "src": ProvidedNode.directory(
                children={
                    "main.py": ProvidedNode.file(content="print('hi')"),
                }
            ),
        }

    vfs.mount_directory("/templates/demo", loader)
    entries = {entry.name for entry in vfs.ls("/templates/demo")}
    assert entries == {"README.md", "src"}
    assert vfs.read_file("/templates/demo/src/main.py") == "print('hi')"


def test_tree_representation(tmp_path):
    vfs = VirtualFileSystem()
    vfs.write_file("/a/b/c.txt", "data")
    tree = vfs.tree("/a")
    assert "c.txt" in tree



"""sandfs package: virtual filesystem sandbox with a mini shell."""

from .vfs import VirtualFileSystem
from .shell import SandboxShell, CommandResult
from .pyexec import PythonExecutor, PythonExecutionResult
from .providers import ContentProvider, DirectoryProvider, NodeContext, ProvidedNode
from .policies import NodePolicy, VisibilityView
from .hooks import WriteEvent, WriteHook
from .integrations import PathEvent, PathHook
from .adapters import MemoryStorageAdapter, StorageAdapter

__all__ = [
    "VirtualFileSystem",
    "SandboxShell",
    "CommandResult",
    "PythonExecutor",
    "PythonExecutionResult",
    "ContentProvider",
    "DirectoryProvider",
    "NodeContext",
    "ProvidedNode",
    "NodePolicy",
    "VisibilityView",
    "WriteEvent",
    "WriteHook",
    "PathEvent",
    "PathHook",
    "StorageAdapter",
    "MemoryStorageAdapter",
]



"""Virtual filesystem implementation."""

from __future__ import annotations

import contextlib
import tempfile
from dataclasses import dataclass
from pathlib import Path, PurePosixPath
from typing import Dict, Iterable, Iterator, List, Optional, Tuple

from .exceptions import InvalidOperation, NodeExists, NodeNotFound
from .hooks import WriteEvent, WriteHook
from .integrations import PathEvent, PathHook
from .nodes import VirtualDirectory, VirtualFile, VirtualNode
from .policies import NodePolicy, VisibilityView
from .providers import ContentProvider, DirectoryProvider
from .adapters import StorageAdapter


@dataclass
class DirEntry:
    name: str
    path: PurePosixPath
    is_dir: bool
    metadata: Dict[str, object]
    policy: NodePolicy


@dataclass
class NodeSnapshot:
    is_dir: bool
    metadata: Dict[str, object]
    policy: NodePolicy
    version: int
    content: Optional[str] = None


@dataclass
class VFSSnapshot:
    nodes: Dict[str, NodeSnapshot]
    cwd: PurePosixPath
    storage_mounts: Dict[str, StorageAdapter]


class VirtualFileSystem:
    """In-memory filesystem that supports dynamic nodes."""

    def __init__(self) -> None:
        self.root = VirtualDirectory(name="")
        self.cwd = self.root
        self._write_hooks: List[Tuple[PurePosixPath, WriteHook]] = []
        self._storage_mounts: Dict[PurePosixPath, StorageAdapter] = {}
        self._path_hooks: List[Tuple[PurePosixPath, PathHook]] = []

    # ------------------------------------------------------------------
    # Path helpers
    # ------------------------------------------------------------------
    def _normalize(self, path: str | PurePosixPath | None) -> PurePosixPath:
        if path is None or str(path) == "":
            base = self.cwd.path()
            raw = base
        else:
            raw = PurePosixPath(path)
            if not raw.is_absolute():
                base = self.cwd.path()
                raw = base.joinpath(raw)
        parts: List[str] = []
        for part in raw.parts:
            if part in ("", "/", "."):
                continue
            if part == "..":
                if parts:
                    parts.pop()
                continue
            parts.append(part)
        return PurePosixPath("/" + "/".join(parts)) if parts else PurePosixPath("/")

    def _iterate_parts(self, path: PurePosixPath) -> Iterable[str]:
        for part in path.parts:
            if part in ("", "/"):
                continue
            yield part

    def _resolve_node(self, path: str | PurePosixPath) -> VirtualNode:
        target = self._normalize(path)
        current: VirtualNode = self.root
        if target == PurePosixPath("/"):
            return self.root
        for part in self._iterate_parts(target):
            if not isinstance(current, VirtualDirectory):
                raise InvalidOperation(f"{current.path()} is not a directory")
            current = current.get_child(part, self)
        return current

    def _resolve_dir(self, path: str | PurePosixPath, *, create: bool = False) -> VirtualDirectory:
        target = self._normalize(path)
        if target == PurePosixPath("/"):
            return self.root
        current = self.root
        for part in self._iterate_parts(target):
            if not isinstance(current, VirtualDirectory):
                raise InvalidOperation(f"{current.path()} is not a directory")
            try:
                next_node = current.get_child(part, self)
            except NodeNotFound:
                if not create:
                    raise
                next_node = VirtualDirectory(name=part, parent=current)
                current.add_child(next_node)
            if not isinstance(next_node, VirtualDirectory):
                raise InvalidOperation(f"{next_node.path()} is not a directory")
            current = next_node
        return current

    def _ensure_file(self, path: str | PurePosixPath, *, create: bool = False) -> VirtualFile:
        target = self._normalize(path)
        parent_path = target.parent
        if parent_path == target:
            raise InvalidOperation("Cannot create file at root path")
        parent = self._resolve_dir(parent_path or PurePosixPath("/"), create=create)
        name = target.name
        if not name:
            raise InvalidOperation("Missing file name")
        try:
            node = parent.get_child(name, self)
        except NodeNotFound:
            if not create:
                raise
            node = VirtualFile(name=name, parent=parent)
            parent.add_child(node)
            return node
        if not isinstance(node, VirtualFile):
            raise InvalidOperation(f"{node.path()} is not a file")
        return node

    def _ensure_read_allowed(self, node: VirtualNode) -> None:
        if not node.policy.readable:
            raise InvalidOperation(f"{node.path()} is not readable")

    def _ensure_write_allowed(self, node: VirtualNode, *, append: bool = False) -> None:
        if not node.policy.writable:
            raise InvalidOperation(f"{node.path()} is read-only")
        if node.policy.append_only and not append:
            raise InvalidOperation(f"{node.path()} is append-only")

    def _check_version(self, node: VirtualNode, expected_version: Optional[int]) -> None:
        if expected_version is None:
            return
        if node.version != expected_version:
            raise InvalidOperation(
                f"Version mismatch for {node.path()}: expected {expected_version}, current {node.version}"
            )

    def _emit_write_event(self, node: VirtualFile, *, append: bool, event_type: str) -> None:
        if self._write_hooks:
            path = node.path()
            content = node.read(self)
            event = WriteEvent(path=str(path), content=content, version=node.version, append=append)
            for prefix, hook in self._write_hooks:
                if self._path_matches_prefix(path, prefix):
                    hook(event)

        self._emit_path_event(node.path(), event_type, node.read(self))

    def _emit_path_event(self, path: PurePosixPath, event_type: str, content: Optional[str]) -> None:
        if not self._path_hooks:
            return
        payload = PathEvent(path=str(path), event=event_type, content=content)
        for prefix, hook in self._path_hooks:
            if self._path_matches_prefix(path, prefix):
                hook(payload)

    def _path_matches_prefix(self, path: PurePosixPath, prefix: PurePosixPath) -> bool:
        if prefix == PurePosixPath("/"):
            return True
        try:
            path.relative_to(prefix)
            return True
        except ValueError:
            return False

    def _clone_policy(self, policy: NodePolicy) -> NodePolicy:
        return NodePolicy(
            readable=policy.readable,
            writable=policy.writable,
            append_only=policy.append_only,
            classification=policy.classification,
            principals=set(policy.principals),
        )

    def _find_storage_mount(
        self, path: PurePosixPath
    ) -> Optional[Tuple[PurePosixPath, StorageAdapter]]:
        matches: List[Tuple[PurePosixPath, StorageAdapter]] = []
        for prefix, adapter in self._storage_mounts.items():
            if self._path_matches_prefix(path, prefix):
                matches.append((prefix, adapter))
        if not matches:
            return None
        return max(matches, key=lambda item: len(item[0].parts))

    def _relative_storage_path(self, path: PurePosixPath, prefix: PurePosixPath) -> str:
        rel = path.relative_to(prefix)
        return rel.as_posix()

    def _persist_storage(self, node: VirtualFile, previous_version: int) -> None:
        mount = self._find_storage_mount(node.path())
        if not mount:
            return
        prefix, adapter = mount
        relative = self._relative_storage_path(node.path(), prefix)
        try:
            adapter.write(relative, node.read(self), version=previous_version)
        except ValueError as exc:
            node.version = previous_version
            raise InvalidOperation(f"Storage conflict for {node.path()}") from exc

    def _delete_storage_entry(self, node: VirtualFile) -> None:
        mount = self._find_storage_mount(node.path())
        if not mount:
            return
        prefix, adapter = mount
        relative = self._relative_storage_path(node.path(), prefix)
        adapter.delete(relative)

    def _load_storage_mount(self, prefix: PurePosixPath, adapter: StorageAdapter) -> None:
        directory = self._resolve_dir(prefix)
        directory.children.clear()
        directory._loaded = True
        for rel_path, entry in adapter.list().items():
            absolute = prefix.joinpath(PurePosixPath(rel_path))
            file_node = self._ensure_file(absolute, create=True)
            file_node.write(entry.content)
            file_node.version = entry.version

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------
    def pwd(self) -> str:
        return str(self.cwd.path())

    def cd(self, path: str | PurePosixPath) -> str:
        node = self._resolve_node(path)
        if not isinstance(node, VirtualDirectory):
            raise InvalidOperation(f"{node.path()} is not a directory")
        self._ensure_read_allowed(node)
        self.cwd = node
        return self.pwd()

    def ls(
        self,
        path: str | PurePosixPath | None = None,
        *,
        view: Optional[VisibilityView] = None,
    ) -> List[DirEntry]:
        directory = self._resolve_dir(path or self.cwd.path())
        self._ensure_read_allowed(directory)
        directory.ensure_loaded(self)
        entries: List[DirEntry] = []
        for child in directory.iter_children(self):
            if view and not view.allows(child.policy):
                continue
            entries.append(
                DirEntry(
                    name=child.name,
                    path=child.path(),
                    is_dir=isinstance(child, VirtualDirectory),
                    metadata=child.metadata,
                    policy=child.policy,
                )
            )
        entries.sort(key=lambda entry: (not entry.is_dir, entry.name))
        return entries

    def mkdir(
        self,
        path: str | PurePosixPath,
        *,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> VirtualDirectory:
        normalized = self._normalize(path)
        if normalized == PurePosixPath("/"):
            return self.root
        parent = self._resolve_dir(normalized.parent or PurePosixPath("/"), create=parents)
        self._ensure_write_allowed(parent)
        name = normalized.name
        if not name:
            raise InvalidOperation("Directory name missing")
        try:
            existing = parent.get_child(name, self)
        except NodeNotFound:
            node = VirtualDirectory(name=name, parent=parent)
            parent.add_child(node)
            return node
        if not isinstance(existing, VirtualDirectory):
            raise InvalidOperation(f"{existing.path()} is not a directory")
        if not exist_ok:
            raise NodeExists(f"Directory {existing.path()} already exists")
        return existing

    def write_file(
        self,
        path: str | PurePosixPath,
        data: str,
        *,
        append: bool = False,
        expected_version: Optional[int] = None,
    ) -> VirtualFile:
        node = self._ensure_file(path, create=True)
        self._ensure_write_allowed(node, append=append)
        self._check_version(node, expected_version)
        previous_version = node.version
        node.write(data, append=append)
        node.version += 1
        self._persist_storage(node, previous_version)
        event_type = "create" if previous_version == 0 else "update"
        self._emit_write_event(node, append=append, event_type=event_type)
        return node

    def append_file(
        self,
        path: str | PurePosixPath,
        data: str,
        *,
        expected_version: Optional[int] = None,
    ) -> VirtualFile:
        return self.write_file(path, data, append=True, expected_version=expected_version)

    def read_file(self, path: str | PurePosixPath) -> str:
        node = self._ensure_file(path, create=False)
        self._ensure_read_allowed(node)
        return node.read(self)

    def touch(self, path: str | PurePosixPath) -> VirtualFile:
        node = self._ensure_file(path, create=True)
        self._ensure_write_allowed(node, append=True)
        return node

    def remove(self, path: str | PurePosixPath, *, recursive: bool = False) -> None:
        target = self._normalize(path)
        if target == PurePosixPath("/"):
            raise InvalidOperation("Cannot remove root directory")
        node = self._resolve_node(target)
        if isinstance(node, VirtualDirectory):
            node.ensure_loaded(self)
        parent = node.parent
        if parent is None:
            raise InvalidOperation("Cannot remove node without parent")
        self._ensure_write_allowed(node)
        self._ensure_write_allowed(parent)
        if isinstance(node, VirtualDirectory) and node.children and not recursive:
            raise InvalidOperation("Directory not empty; pass recursive=True")
        if isinstance(node, VirtualDirectory) and recursive:
            names = list(node.children.keys())
            for child_name in names:
                child_node = node.children.get(child_name)
                if child_node is None:
                    continue
                self.remove(child_node.path(), recursive=True)
        parent.remove_child(node.name)
        if isinstance(node, VirtualFile):
            self._delete_storage_entry(node)
            self._emit_path_event(node.path(), "delete", None)

    def walk(self, path: str | PurePosixPath | None = None) -> Iterator[Tuple[PurePosixPath, VirtualNode]]:
        start_node = self._resolve_node(path or self.cwd.path())

        def _walk(node: VirtualNode) -> Iterator[Tuple[PurePosixPath, VirtualNode]]:
            yield (node.path(), node)
            if isinstance(node, VirtualDirectory):
                node.ensure_loaded(self)
                for child in node.iter_children(self):
                    yield from _walk(child)

        return _walk(start_node)

    def iter_files(
        self,
        path: str | PurePosixPath | None = None,
        *,
        recursive: bool = True,
    ) -> Iterator[Tuple[PurePosixPath, VirtualFile]]:
        start_node = self._resolve_node(path or self.cwd.path())
        self._ensure_read_allowed(start_node)
        if isinstance(start_node, VirtualFile):
            yield (start_node.path(), start_node)
            return

        directory = self._resolve_dir(start_node.path())
        directory.ensure_loaded(self)

        def _walk_dir(dir_node: VirtualDirectory) -> Iterator[Tuple[PurePosixPath, VirtualFile]]:
            for child in dir_node.iter_children(self):
                if isinstance(child, VirtualFile):
                    yield (child.path(), child)
                elif isinstance(child, VirtualDirectory) and recursive:
                    child.ensure_loaded(self)
                    yield from _walk_dir(child)

        yield from _walk_dir(directory)

    def snapshot(self) -> VFSSnapshot:
        nodes: Dict[str, NodeSnapshot] = {}
        for path, node in self.walk("/"):
            if isinstance(node, VirtualFile):
                content = node.read(self)
            else:
                content = None
            nodes[str(path)] = NodeSnapshot(
                is_dir=isinstance(node, VirtualDirectory),
                metadata=dict(node.metadata),
                policy=self._clone_policy(node.policy),
                version=node.version,
                content=content,
            )
        storage_mounts = {str(path): adapter for path, adapter in self._storage_mounts.items()}
        return VFSSnapshot(nodes=nodes, cwd=self.cwd.path(), storage_mounts=storage_mounts)

    def restore(self, snapshot: VFSSnapshot) -> None:
        self.root = VirtualDirectory(name="")
        self.cwd = self.root
        self._storage_mounts = {PurePosixPath(path): adapter for path, adapter in snapshot.storage_mounts.items()}
        ordered = sorted(snapshot.nodes.items(), key=lambda item: len(PurePosixPath(item[0]).parts))
        for path_str, node_state in ordered:
            path = PurePosixPath(path_str)
            if path == PurePosixPath("/"):
                target = self.root
                target.metadata = dict(node_state.metadata)
                target.policy = self._clone_policy(node_state.policy)
                target.version = node_state.version
                continue
            if node_state.is_dir:
                directory = self.mkdir(path, parents=True, exist_ok=True)
                directory.metadata = dict(node_state.metadata)
                directory.policy = self._clone_policy(node_state.policy)
                directory.version = node_state.version
            else:
                file_node = self._ensure_file(path, create=True)
                file_node.metadata = dict(node_state.metadata)
                file_node.policy = self._clone_policy(node_state.policy)
                file_node.write(node_state.content or "")
                file_node.version = node_state.version
        self.cwd = self._resolve_dir(snapshot.cwd)
    def export_to_path(
        self,
        target: Path,
        *,
        source: str | PurePosixPath | None = None,
    ) -> Path:
        node = self._resolve_dir(source or self.cwd.path())
        target = Path(target)
        target.mkdir(parents=True, exist_ok=True)
        self._export_directory(node, target)
        return target

    @contextlib.contextmanager
    def materialize(self, path: str | PurePosixPath | None = None):
        with tempfile.TemporaryDirectory() as tmp:
            root = Path(tmp)
            self.export_to_path(root, source=path)
            yield root

    def _export_directory(self, node: VirtualDirectory, dest: Path) -> None:
        node.ensure_loaded(self)
        dest.mkdir(parents=True, exist_ok=True)
        for child in node.iter_children(self):
            target = dest / child.name
            if isinstance(child, VirtualDirectory):
                self._export_directory(child, target)
            else:
                target.parent.mkdir(parents=True, exist_ok=True)
                target.write_text(child.read(self))

    def exists(self, path: str | PurePosixPath) -> bool:
        try:
            self._resolve_node(path)
            return True
        except (NodeNotFound, InvalidOperation):
            return False

    def get_version(self, path: str | PurePosixPath) -> int:
        node = self._resolve_node(path)
        if not isinstance(node, VirtualFile):
            raise InvalidOperation(f"{node.path()} is not a file")
        return node.version

    def is_dir(self, path: str | PurePosixPath) -> bool:
        try:
            return isinstance(self._resolve_node(path), VirtualDirectory)
        except (NodeNotFound, InvalidOperation):
            return False

    def is_file(self, path: str | PurePosixPath) -> bool:
        try:
            return isinstance(self._resolve_node(path), VirtualFile)
        except (NodeNotFound, InvalidOperation):
            return False

    def mount_file(
        self,
        path: str | PurePosixPath,
        provider: ContentProvider,
        *,
        metadata: Optional[Dict[str, object]] = None,
    ) -> VirtualFile:
        node = self._ensure_file(path, create=True)
        node.set_provider(provider)
        if metadata:
            node.metadata.update(metadata)
        return node

    def mount_directory(
        self,
        path: str | PurePosixPath,
        provider: DirectoryProvider,
        *,
        metadata: Optional[Dict[str, object]] = None,
    ) -> VirtualDirectory:
        node = self.mkdir(path, parents=True, exist_ok=True)
        node.loader = provider
        node._loaded = False  # allow reload
        if metadata:
            node.metadata.update(metadata)
        return node

    def mount_storage(
        self,
        path: str | PurePosixPath,
        adapter: StorageAdapter,
        *,
        policy: Optional[NodePolicy] = None,
    ) -> VirtualDirectory:
        normalized = self._normalize(path)
        directory = self.mkdir(normalized, parents=True, exist_ok=True)
        if policy is not None:
            directory.policy = policy
        self._storage_mounts[normalized] = adapter
        self._load_storage_mount(normalized, adapter)
        return directory

    def sync_storage(self, path: str | PurePosixPath) -> None:
        normalized = self._normalize(path)
        adapter = self._storage_mounts.get(normalized)
        if adapter is None:
            raise InvalidOperation(f"No storage mount at {normalized}")
        self._load_storage_mount(normalized, adapter)

    def register_write_hook(self, prefix: str | PurePosixPath, hook: WriteHook) -> None:
        normalized = self._normalize(prefix)
        self._write_hooks.append((normalized, hook))

    def register_path_hook(self, prefix: str | PurePosixPath, hook: PathHook) -> None:
        normalized = self._normalize(prefix)
        self._path_hooks.append((normalized, hook))

    def set_policy(self, path: str | PurePosixPath, policy: NodePolicy) -> None:
        node = self._resolve_node(path)
        node.policy = policy

    def get_policy(self, path: str | PurePosixPath) -> NodePolicy:
        node = self._resolve_node(path)
        return node.policy

    def get_node(self, path: str | PurePosixPath) -> VirtualNode:
        return self._resolve_node(path)

    def tree(
        self,
        path: str | PurePosixPath | None = None,
        *,
        view: Optional[VisibilityView] = None,
    ) -> str:
        root_dir = self._resolve_dir(path or self.cwd.path())
        self._ensure_read_allowed(root_dir)
        lines: List[str] = []

        def render(directory: VirtualDirectory, prefix: str = "") -> None:
            entries = sorted(
                (
                    child
                    for child in directory.iter_children(self)
                    if not view or view.allows(child.policy)
                ),
                key=lambda node: (not isinstance(node, VirtualDirectory), node.name),
            )
            for idx, node in enumerate(entries):
                connector = "└──" if idx == len(entries) - 1 else "├──"
                lines.append(f"{prefix}{connector} {node.name}/" if isinstance(node, VirtualDirectory) else f"{prefix}{connector} {node.name}")
                if isinstance(node, VirtualDirectory):
                    extension = "    " if idx == len(entries) - 1 else "│   "
                    render(node, prefix + extension)

        render(root_dir)
        header = str(root_dir.path())
        return "\n".join([header] + lines)


__all__ = ["VirtualFileSystem", "DirEntry"]



import pytest

from sandfs import NodePolicy, VirtualFileSystem
from sandfs.adapters import MemoryStorageAdapter
from sandfs.exceptions import InvalidOperation


def test_storage_adapter_mount_and_persist():
    adapter = MemoryStorageAdapter(
        initial={
            "a.txt": "hello",
            "dir/b.txt": "nested",
        }
    )
    vfs = VirtualFileSystem()
    vfs.mount_storage("/data", adapter, policy=NodePolicy(writable=True))

    assert vfs.read_file("/data/a.txt") == "hello"
    vfs.write_file("/data/a.txt", "world")
    vfs.write_file("/data/dir/new.txt", "fresh")

    assert adapter.read("a.txt").content == "world"
    assert adapter.read("dir/new.txt").content == "fresh"


def test_storage_adapter_conflict_detection():
    adapter = MemoryStorageAdapter(initial={"a.txt": "hello"})
    vfs = VirtualFileSystem()
    vfs.mount_storage("/logs", adapter)

    vfs.write_file("/logs/a.txt", "one")
    current = adapter.read("a.txt")
    adapter.write("a.txt", "external", version=current.version)

    with pytest.raises(InvalidOperation):
        vfs.write_file("/logs/a.txt", "two", expected_version=1)


def test_storage_adapter_sync_refreshes_vfs():
    adapter = MemoryStorageAdapter(initial={"a.txt": "hello"})
    vfs = VirtualFileSystem()
    vfs.mount_storage("/sync", adapter)

    current = adapter.read("a.txt")
    adapter.write("a.txt", "external", version=current.version)
    vfs.sync_storage("/sync")
    assert vfs.read_file("/sync/a.txt") == "external"



"""Pure-Python shell facade around the virtual filesystem."""

from __future__ import annotations

import re
import shlex
import subprocess
from dataclasses import dataclass
from typing import Callable, Dict, Iterable, List, Optional, Set

from .exceptions import InvalidOperation, SandboxError, NodeNotFound
from .policies import VisibilityView
from .pyexec import PythonExecutor
from .vfs import DirEntry, VirtualFileSystem


@dataclass
class CommandResult:
    stdout: str = ""
    stderr: str = ""
    exit_code: int = 0


CommandHandler = Callable[[List[str]], CommandResult | str | None]


class SandboxShell:
    """Executes a curated subset of shell commands against the VFS."""

    def __init__(
        self,
        vfs: VirtualFileSystem,
        *,
        python_executor: Optional[PythonExecutor] = None,
        env: Optional[Dict[str, str]] = None,
        view: Optional[VisibilityView] = None,
        allowed_commands: Optional[Iterable[str]] = None,
        max_output_bytes: Optional[int] = None,
        host_fallback: bool = True,
    ) -> None:
        self.vfs = vfs
        self.env: Dict[str, str] = dict(env or {})
        self.commands: Dict[str, CommandHandler] = {}
        self.command_docs: Dict[str, str] = {}
        self.last_command_name: Optional[str] = None
        self.command_docs: Dict[str, str] = {}
        self.py_exec = python_executor or PythonExecutor(vfs)
        self.view = view or VisibilityView()
        self.allowed_commands: Optional[Set[str]] = set(allowed_commands) if allowed_commands else None
        self.max_output_bytes = max_output_bytes
        self.host_fallback = host_fallback
        self._register_builtin_commands()

    # ------------------------------------------------------------------
    # Command registration
    # ------------------------------------------------------------------
    def register_command(self, name: str, handler: CommandHandler, *, description: str = "") -> None:
        self.commands[name] = handler
        if description:
            self.command_docs[name] = description

    def available_commands(self) -> List[str]:
        return sorted(self.commands)

    def _ensure_visible_path(self, path: str) -> None:
        if self.view is None:
            return
        try:
            policy = self.vfs.get_policy(path)
        except NodeNotFound:
            return
        if not self.view.allows(policy):
            raise InvalidOperation(f"Path {path} is hidden for this view")

    def _enforce_output_limit(self, result: CommandResult) -> CommandResult:
        if self.max_output_bytes is None:
            return result
        total = len(result.stdout) + len(result.stderr)
        if total <= self.max_output_bytes:
            return result
        return CommandResult(
            stdout="",
            stderr=f"Output limit ({self.max_output_bytes} bytes) exceeded",
            exit_code=1,
        )

    def _run_host_process(self, command_tokens: List[str], path: Optional[str]) -> CommandResult:
        if not command_tokens:
            return CommandResult(stderr="Missing host command", exit_code=2)
        target = path or self.vfs.pwd()
        self._ensure_visible_path(target)
        try:
            with self.vfs.materialize(target) as fs_root:
                completed = subprocess.run(
                    command_tokens,
                    cwd=str(fs_root),
                    capture_output=True,
                    text=True,
                    check=False,
                )
        except SandboxError as exc:
            return CommandResult(stderr=str(exc), exit_code=1)
        except FileNotFoundError as exc:
            return CommandResult(stderr=str(exc), exit_code=127)
        return CommandResult(
            stdout=completed.stdout,
            stderr=completed.stderr,
            exit_code=completed.returncode,
        )

    # ------------------------------------------------------------------
    # Dispatcher
    # ------------------------------------------------------------------
    def exec(self, command: str) -> CommandResult:
        last_result = CommandResult()
        for segment in filter(None, (line.strip() for line in command.splitlines())):
            last_result = self._exec_one(segment)
            if last_result.exit_code != 0:
                return last_result
        return last_result

    def _exec_one(self, command: str) -> CommandResult:
        if not command.strip():
            return CommandResult()
        try:
            tokens = shlex.split(command)
        except ValueError as exc:
            return CommandResult(stderr=str(exc), exit_code=2)
        if not tokens:
            return CommandResult()
        name, *args = tokens
        self.last_command_name = name
        handler = self.commands.get(name)
        if handler is None:
            if self.host_fallback:
                return self._run_host_process(tokens, None)
            return CommandResult(stderr=f"Unknown command: {name}", exit_code=127)
        if self.allowed_commands is not None and name not in self.allowed_commands:
            return CommandResult(stderr=f"Command '{name}' is disabled in this shell", exit_code=1)
        try:
            result = handler(args)
        except SandboxError as exc:
            return CommandResult(stderr=str(exc), exit_code=1)
        except Exception as exc:  # pragma: no cover - unexpected failure path
            return CommandResult(stderr=f"{name} failed: {exc}", exit_code=1)
        if isinstance(result, CommandResult):
            return self._enforce_output_limit(result)
        if result is None:
            return self._enforce_output_limit(CommandResult())
        return self._enforce_output_limit(CommandResult(stdout=str(result)))

    # ------------------------------------------------------------------
    # Builtin commands
    # ------------------------------------------------------------------
    def _register_builtin_commands(self) -> None:
        self.register_command("pwd", self._cmd_pwd, description="Print working directory")
        self.register_command("cd", self._cmd_cd, description="Change directory")
        self.register_command("ls", self._cmd_ls, description="List directory contents")
        self.register_command("cat", self._cmd_cat, description="Print file contents")
        self.register_command("touch", self._cmd_touch, description="Create empty file")
        self.register_command("mkdir", self._cmd_mkdir, description="Create directories")
        self.register_command("rm", self._cmd_rm, description="Remove files or directories")
        self.register_command("tree", self._cmd_tree, description="Render tree view")
        self.register_command("write", self._cmd_write, description="Write text to file")
        self.register_command("append", self._cmd_append, description="Append text to file")
        self.register_command("grep", self._cmd_grep, description="Search files (non-recursive)")
        self.register_command("rg", self._cmd_rg, description="Search files recursively")
        self.register_command("python", self._cmd_python, description="Execute Python snippet")
        self.register_command("python3", self._cmd_python, description="Execute Python snippet")
        self.register_command("host", self._cmd_host, description="Run host command in materialized tree")
        self.register_command("bash", self._cmd_shell_host, description="Run bash via host")
        self.register_command("sh", self._cmd_shell_host, description="Run sh via host")
        self.register_command("help", self._cmd_help, description="Show available commands")

    def _cmd_pwd(self, _: List[str]) -> CommandResult:
        return CommandResult(stdout=self.vfs.pwd())

    def _cmd_cd(self, args: List[str]) -> CommandResult:
        if len(args) != 1:
            return CommandResult(stderr="cd expects exactly one path", exit_code=2)
        self._ensure_visible_path(args[0])
        new_path = self.vfs.cd(args[0])
        return CommandResult(stdout=new_path)

    def _cmd_ls(self, args: List[str]) -> CommandResult:
        long = False
        targets: List[str] = []
        for arg in args:
            if arg in ("-l", "--long"):
                long = True
            elif arg.startswith("-"):
                # fallback for other flags via host
                return self._run_host_process(["ls", *args], None)
            else:
                targets.append(arg)
        if not targets:
            targets = [self.vfs.pwd()]
        blocks: List[str] = []
        for idx, target in enumerate(targets):
            self._ensure_visible_path(target)
            entries = self.vfs.ls(target, view=self.view)
            if len(targets) > 1:
                blocks.append(f"{target}:")
            blocks.append(self._format_ls(entries, long_format=long))
            if idx < len(targets) - 1:
                blocks.append("")
        return CommandResult(stdout="\n".join(filter(None, blocks)))

    def _format_ls(self, entries: List[DirEntry], *, long_format: bool) -> str:
        if not entries:
            return ""
        if long_format:
            return "\n".join(
                f"{'d' if entry.is_dir else '-'} {entry.path}"
                for entry in entries
            )
        return "  ".join(
            f"{entry.name}/" if entry.is_dir else entry.name
            for entry in entries
        )

    def _cmd_cat(self, args: List[str]) -> CommandResult:
        if not args:
            return CommandResult(stderr="cat expects at least one file", exit_code=2)
        blobs = []
        for path in args:
            self._ensure_visible_path(path)
            blobs.append(self.vfs.read_file(path))
        return CommandResult(stdout="".join(blobs))

    def _cmd_append(self, args: List[str]) -> CommandResult:
        if len(args) < 2:
            return CommandResult(stderr="append expects a path and text", exit_code=2)
        path = args[0]
        self._ensure_visible_path(path)
        text = " ".join(args[1:])
        self.vfs.append_file(path, text)
        return CommandResult()

    def _cmd_touch(self, args: List[str]) -> CommandResult:
        if not args:
            return CommandResult(stderr="touch expects at least one file", exit_code=2)
        for path in args:
            self._ensure_visible_path(path)
            self.vfs.touch(path)
        return CommandResult()

    def _cmd_mkdir(self, args: List[str]) -> CommandResult:
        parents = False
        paths: List[str] = []
        for arg in args:
            if arg in ("-p", "--parents"):
                parents = True
            else:
                paths.append(arg)
        if not paths:
            return CommandResult(stderr="mkdir expects a path", exit_code=2)
        for path in paths:
            self._ensure_visible_path(path)
            self.vfs.mkdir(path, parents=parents, exist_ok=parents)
        return CommandResult()

    def _cmd_rm(self, args: List[str]) -> CommandResult:
        recursive = False
        targets: List[str] = []
        for arg in args:
            if arg in ("-r", "-rf", "-R", "--recursive"):
                recursive = True
            else:
                targets.append(arg)
        if not targets:
            return CommandResult(stderr="rm expects a target", exit_code=2)
        for target in targets:
            self._ensure_visible_path(target)
            self.vfs.remove(target, recursive=recursive)
        return CommandResult()

    def _cmd_tree(self, args: List[str]) -> CommandResult:
        target = args[0] if args else None
        if target:
            self._ensure_visible_path(target)
        return CommandResult(stdout=self.vfs.tree(target, view=self.view))

    def _cmd_write(self, args: List[str]) -> CommandResult:
        if not args:
            return CommandResult(stderr="write expects a target path", exit_code=2)
        path = args[0]
        self._ensure_visible_path(path)
        text_parts: List[str] = []
        append = False
        idx = 1
        while idx < len(args):
            token = args[idx]
            if token == "--append":
                append = True
                idx += 1
                continue
            if token == "--text" and idx + 1 < len(args):
                text_parts.append(args[idx + 1])
                idx += 2
                continue
            text_parts.append(token)
            idx += 1
        payload = " ".join(text_parts)
        if append:
            self.vfs.append_file(path, payload)
        else:
            self.vfs.write_file(path, payload)
        return CommandResult()

    def _cmd_grep(self, args: List[str]) -> CommandResult:
        if not args:
            return CommandResult(stderr="grep expects a pattern", exit_code=2)
        recursive = False
        regex = False
        ignore_case = False
        show_numbers = False
        paths: List[str] = []
        pattern: Optional[str] = None
        idx = 0
        while idx < len(args):
            token = args[idx]
            if token in ("-r", "-R", "--recursive"):
                recursive = True
                idx += 1
                continue
            if token in ("-i", "--ignore-case"):
                ignore_case = True
                idx += 1
                continue
            if token in ("-n", "--line-number"):
                show_numbers = True
                idx += 1
                continue
            if token in ("-e", "--regex"):
                regex = True
                idx += 1
                continue
            if pattern is None:
                pattern = token
            else:
                paths.append(token)
            idx += 1
        if pattern is None:
            return CommandResult(stderr="Missing pattern", exit_code=2)
        if not paths:
            paths = [self.vfs.pwd()]
        for target in paths:
            self._ensure_visible_path(target)
        output = self._search(pattern, paths, recursive=recursive, regex=regex, ignore_case=ignore_case, show_numbers=show_numbers)
        return CommandResult(stdout="\n".join(output))

    def _cmd_rg(self, args: List[str]) -> CommandResult:
        # ripgrep defaults to recursive search
        return self._cmd_grep(["-r"] + args)

    def _search(
        self,
        pattern: str,
        paths: Iterable[str],
        *,
        recursive: bool,
        regex: bool,
        ignore_case: bool,
        show_numbers: bool,
    ) -> List[str]:
        results: List[str] = []
        flags = re.MULTILINE
        if ignore_case:
            flags |= re.IGNORECASE
        compiled = re.compile(pattern, flags) if regex else None
        lowered = pattern.lower() if ignore_case and not regex else None
        for target in paths:
            for file_path, file_node in self.vfs.iter_files(target, recursive=recursive):
                if self.view and not self.view.allows(file_node.policy):
                    continue
                text = file_node.read(self.vfs)
                lines = text.splitlines()
                for idx, line in enumerate(lines, start=1):
                    matched = False
                    if regex:
                        if compiled and compiled.search(line):
                            matched = True
                    elif ignore_case:
                        if lowered and lowered in line.lower():
                            matched = True
                    else:
                        if pattern in line:
                            matched = True
                    if matched:
                        prefix = f"{file_path}:{idx}:" if show_numbers else f"{file_path}:"
                        results.append(f"{prefix}{line}")
        return results

    def _cmd_python(self, args: List[str]) -> CommandResult:
        if not args:
            return CommandResult(stderr="python expects code", exit_code=2)
        if args[0] == "-c" and len(args) >= 2:
            code = " ".join(args[1:])
        else:
            code = " ".join(args)
        result = self.py_exec.run(code)
        return CommandResult(stdout=result.stdout)

    def _cmd_shell_host(self, args: List[str]) -> CommandResult:
        # First token is bash/sh command itself; delegate to host with same args
        return self._run_host_process([self.last_command_name] + args, None)

    def _cmd_host(self, args: List[str]) -> CommandResult:
        path: Optional[str] = None
        idx = 0
        while idx < len(args):
            token = args[idx]
            if token in ("-p", "--path", "-C"):
                idx += 1
                if idx >= len(args):
                    return CommandResult(stderr="host expects a path after -p/--path", exit_code=2)
                path = args[idx]
                idx += 1
                continue
            if token == "--":
                idx += 1
                break
            break
        command_tokens = args[idx:]
        if not command_tokens:
            return CommandResult(stderr="host expects a command to run", exit_code=2)
        return self._run_host_process(command_tokens, path)

    def _cmd_help(self, _: List[str]) -> CommandResult:
        lines = ["Available commands:"]
        for name in self.available_commands():
            desc = self.command_docs.get(name, "")
            if desc:
                lines.append(f"  {name} - {desc}")
            else:
                lines.append(f"  {name}")
        lines.append("Use host <cmd> (or run unknown commands directly) for full GNU tools.")
        return CommandResult(stdout="\n".join(lines))


__all__ = ["SandboxShell", "CommandResult"]
